@@ -13,8 +13,8 @@
  *                                                                         *
  ***************************************************************************/
 
-#ifndef QGSRULEBASEDRENDERERV2_H
-#define QGSRULEBASEDRENDERERV2_H
+#ifndef QGSRULEBASEDRENDERER_H
+#define QGSRULEBASEDRENDERER_H
 
 #include "qgis_core.h"
 #include "qgis_sip.h"
@@ -23,6 +23,7 @@
 #include "qgis.h"
 
 #include "qgsrenderer.h"
+#include "qgsrendercontext.h"
 
 class QgsExpression;
 
@@ -31,8 +32,10 @@ class QgsGraduatedSymbolRenderer;
 
 /**
  * \ingroup core
-When drawing a vector layer with rule-based renderer, it goes through
-the rules and draws features with symbols from rules that match.
+ * \brief Rule based renderer.
+ *
+ * When drawing a vector layer with rule-based renderer, it goes through
+ * the rules and draws features with symbols from rules that match.
  */
 class CORE_EXPORT QgsRuleBasedRenderer : public QgsFeatureRenderer
 {
@@ -65,16 +68,27 @@ class CORE_EXPORT QgsRuleBasedRenderer : public QgsFeatureRenderer
      */
     struct RenderJob
     {
-      RenderJob( QgsRuleBasedRenderer::FeatureToRender &_ftr, QgsSymbol *_s )
-        : ftr( _ftr )
-        , symbol( _s )
-      {}
 
-      //! Feature to render
-      QgsRuleBasedRenderer::FeatureToRender &ftr;
+        /**
+         * Constructor for a render job, with the specified feature to render and symbol.
+         *
+         * \note The symbol ownership is not transferred.
+         */
+        RenderJob( const QgsRuleBasedRenderer::FeatureToRender &_ftr, QgsSymbol *_s )
+          : ftr( _ftr )
+          , symbol( _s )
+        {}
 
-      //! Symbol to render feature with (not owned by this object).
-      QgsSymbol *symbol = nullptr;
+        //! Feature to render
+        QgsRuleBasedRenderer::FeatureToRender ftr;
+
+        //! Symbol to render feature with (not owned by this object).
+        QgsSymbol *symbol = nullptr;
+
+      private:
+#ifdef SIP_RUN
+        RenderJob &operator=( const RenderJob & );
+#endif
     };
 
     /**
@@ -83,7 +97,10 @@ class CORE_EXPORT QgsRuleBasedRenderer : public QgsFeatureRenderer
      */
     struct RenderLevel
     {
-      explicit RenderLevel( int z ): zIndex( z ) {}
+      explicit RenderLevel( int z )
+        : zIndex( z )
+      {}
+
       ~RenderLevel() { qDeleteAll( jobs ); }
       int zIndex;
 
@@ -95,19 +112,19 @@ class CORE_EXPORT QgsRuleBasedRenderer : public QgsFeatureRenderer
         zIndex = rh.zIndex;
         qDeleteAll( jobs );
         jobs.clear();
-        for ( RenderJob *job :  qgis::as_const( rh.jobs ) )
+        for ( auto it = rh.jobs.constBegin(); it != rh.jobs.constEnd(); ++it )
         {
-          jobs << new RenderJob( *job );
+          jobs << new RenderJob( *( *it ) );
         }
         return *this;
       }
 
       RenderLevel( const QgsRuleBasedRenderer::RenderLevel &other )
-        : zIndex( other.zIndex )
+        : zIndex( other.zIndex ), jobs()
       {
-        for ( RenderJob *job : qgis::as_const( other.jobs ) )
+        for ( auto it = other.jobs.constBegin(); it != other.jobs.constEnd(); ++it )
         {
-          jobs << new RenderJob( *job );
+          jobs << new RenderJob( * ( *it ) );
         }
       }
 
@@ -121,12 +138,13 @@ class CORE_EXPORT QgsRuleBasedRenderer : public QgsFeatureRenderer
 
     /**
      * \ingroup core
-      This class keeps data about a rules for rule-based renderer.
-      A rule consists of a symbol, filter expression and range of scales.
-      If filter is empty, it matches all features.
-      If scale range has both values zero, it matches all scales.
-      If one of the min/max scale denominators is zero, there is no lower/upper bound for scales.
-      A rule matches if both filter and scale range match.
+     * \brief  This class keeps data about a rules for rule-based renderer.
+     *
+     * A rule consists of a symbol, filter expression and range of scales.
+     * If filter is empty, it matches all features.
+     * If scale range has both values zero, it matches all scales.
+     * If one of the min/max scale denominators is zero, there is no lower/upper bound for scales.
+     * A rule matches if both filter and scale range match.
      */
     class CORE_EXPORT Rule
     {
@@ -163,14 +181,13 @@ class CORE_EXPORT QgsRuleBasedRenderer : public QgsFeatureRenderer
         QSet<QString> usedAttributes( const QgsRenderContext &context ) const;
 
         /**
-         * Returns true if this rule or one of its chilren needs the geometry to be applied.
+         * Returns TRUE if this rule or one of its children needs the geometry to be applied.
          */
         bool needsGeometry() const;
 
         //! \note available in Python bindings as symbol2
         QgsSymbolList symbols( const QgsRenderContext &context = QgsRenderContext() ) const;
 
-        //! \since QGIS 2.6
         QgsLegendSymbolList legendSymbolItems( int currentLevel = -1 ) const;
 
         /**
@@ -178,14 +195,14 @@ class CORE_EXPORT QgsRuleBasedRenderer : public QgsFeatureRenderer
          *
          * \param f         The feature to test
          * \param context   The context in which the rendering happens
-         * \returns          True if the feature shall be rendered
+         * \returns          TRUE if the feature shall be rendered
          */
         bool isFilterOK( const QgsFeature &f, QgsRenderContext *context = nullptr ) const;
 
         /**
          * Check if this rule applies for a given \a scale.
          * The \a scale value indicates the scale denominator, e.g. 1000.0 for a 1:1000 map.
-         * If set to 0, it will always return true.
+         * If set to 0, it will always return TRUE.
          *
          * \returns If the rule will be evaluated at this scale
          */
@@ -201,7 +218,6 @@ class CORE_EXPORT QgsRuleBasedRenderer : public QgsFeatureRenderer
          * A scale of 0 indicates no maximum scale visibility.
          * \see minimumScale()
          * \see setMaximumScale()
-         * \since QGIS 3.0
         */
         double maximumScale() const { return mMaximumScale; }
 
@@ -211,7 +227,6 @@ class CORE_EXPORT QgsRuleBasedRenderer : public QgsFeatureRenderer
          * A scale of 0 indicates no minimum scale visibility.
          * \see maximumScale()
          * \see setMinimumScale()
-         * \since QGIS 3.0
         */
         double minimumScale() const { return mMinimumScale; }
 
@@ -237,23 +252,21 @@ class CORE_EXPORT QgsRuleBasedRenderer : public QgsFeatureRenderer
         /**
          * Returns if this rule is active
          *
-         * \returns True if the rule is active
+         * \returns TRUE if the rule is active
          */
         bool active() const { return mIsActive; }
 
         /**
          * Unique rule identifier (for identification of rule within renderer)
-         * \since QGIS 2.6
          */
         QString ruleKey() const { return mRuleKey; }
 
         /**
          * Override the assigned rule key (should be used just internally by rule-based renderer)
-         * \since QGIS 2.6
          */
         void setRuleKey( const QString &key ) { mRuleKey = key; }
 
-        //! Sets a new symbol (or NULL). Deletes old symbol.
+        //! Sets a new symbol (or NULLPTR). Deletes old symbol.
         void setSymbol( QgsSymbol *sym SIP_TRANSFER );
         void setLabel( const QString &label ) { mLabel = label; }
 
@@ -298,18 +311,28 @@ class CORE_EXPORT QgsRuleBasedRenderer : public QgsFeatureRenderer
         //! clone this rule, return new instance
         QgsRuleBasedRenderer::Rule *clone() const SIP_FACTORY;
 
-        void toSld( QDomDocument &doc, QDomElement &element, QgsStringMap props ) const;
+        //! Saves the symbol layer as SLD
+        void toSld( QDomDocument &doc, QDomElement &element, QVariantMap props ) const;
 
         /**
          * Create a rule from the SLD provided in element and for the specified geometry type.
          */
-        static QgsRuleBasedRenderer::Rule *createFromSld( QDomElement &element, QgsWkbTypes::GeometryType geomType ) SIP_FACTORY;
+        static QgsRuleBasedRenderer::Rule *createFromSld( QDomElement &element, Qgis::GeometryType geomType ) SIP_FACTORY;
 
         QDomElement save( QDomDocument &doc, QgsSymbolMap &symbolMap ) const;
 
         //! prepare the rule for rendering and its children (build active children array)
         bool startRender( QgsRenderContext &context, const QgsFields &fields, QString &filter );
 
+        /**
+         * Returns TRUE if the rule has any active children.
+         *
+         * \note Must be called after startRender()
+         *
+         * \since QGIS 3.30
+         */
+        bool hasActiveChildren() const;
+
         //! Gets all used z-levels from this rule and children
         QSet<int> collectZLevels();
 
@@ -338,7 +361,6 @@ class CORE_EXPORT QgsRuleBasedRenderer : public QgsFeatureRenderer
 
         /**
          * Returns which legend keys match the feature
-         * \since QGIS 2.14
          */
         QSet< QString > legendKeysForFeature( const QgsFeature &feature, QgsRenderContext *context = nullptr );
 
@@ -348,7 +370,7 @@ class CORE_EXPORT QgsRuleBasedRenderer : public QgsFeatureRenderer
          *
          * \param feature The feature for which rules have to be find
          * \param context The rendering context
-         * \param onlyActive True to search for active rules only, false otherwise
+         * \param onlyActive TRUE to search for active rules only, FALSE otherwise
          */
         QgsRuleBasedRenderer::RuleList rulesForFeature( const QgsFeature &feature, QgsRenderContext *context = nullptr, bool onlyActive = true );
 
@@ -364,17 +386,18 @@ class CORE_EXPORT QgsRuleBasedRenderer : public QgsFeatureRenderer
          *
          * \param ruleElem  The XML rule element
          * \param symbolMap Symbol map
+         * \param reuseId set to TRUE to create an exact copy of the original symbol or FALSE to create a new rule with the same parameters as the original but a new unique ruleKey(). (Since QGIS 3.30)
          *
          * \returns A new rule
          */
-        static QgsRuleBasedRenderer::Rule *create( QDomElement &ruleElem, QgsSymbolMap &symbolMap ) SIP_FACTORY;
+        static QgsRuleBasedRenderer::Rule *create( QDomElement &ruleElem, QgsSymbolMap &symbolMap, bool reuseId = true ) SIP_FACTORY;
 
         /**
          * Returns all children rules of this rule
          *
          * \returns A list of rules
          */
-        const QgsRuleBasedRenderer::RuleList &children() { return mChildren; }
+        const QgsRuleBasedRenderer::RuleList &children() const { return mChildren; }
 
         /**
          * Returns all children, grand-children, grand-grand-children, grand-gra... you get it
@@ -402,32 +425,42 @@ class CORE_EXPORT QgsRuleBasedRenderer : public QgsFeatureRenderer
         //! delete child rule
         void removeChildAt( int i );
 
-        //! take child rule out, set parent as null
+        //! take child rule out, set parent as NULLPTR
         QgsRuleBasedRenderer::Rule *takeChild( QgsRuleBasedRenderer::Rule *rule ) SIP_TRANSFERBACK;
 
-        //! take child rule out, set parent as null
+        //! take child rule out, set parent as NULLPTR
         QgsRuleBasedRenderer::Rule *takeChildAt( int i ) SIP_TRANSFERBACK;
 
         /**
          * Try to find a rule given its unique key
-         * \since QGIS 2.6
          */
         QgsRuleBasedRenderer::Rule *findRuleByKey( const QString &key );
 
         /**
          * Sets if this rule is an ELSE rule
          *
-         * \param iselse If true, this rule is an ELSE rule
+         * \param iselse If TRUE, this rule is an ELSE rule
          */
         void setIsElse( bool iselse );
 
         /**
          * Check if this rule is an ELSE rule
          *
-         * \returns True if this rule is an else rule
+         * \returns TRUE if this rule is an else rule
          */
         bool isElse() const { return mElseRule; }
 
+        /**
+         * Accepts the specified symbology \a visitor, causing it to visit all child rules associated
+         * with the rule.
+         *
+         * Returns TRUE if the visitor should continue visiting other objects, or FALSE if visiting
+         * should be canceled.
+         *
+         * \since QGIS 3.10
+         */
+        bool accept( QgsStyleEntityVisitorInterface *visitor ) const;
+
       protected:
         void initFilter();
 
@@ -436,7 +469,7 @@ class CORE_EXPORT QgsRuleBasedRenderer : public QgsFeatureRenderer
         Rule( const QgsRuleBasedRenderer::Rule &rh );
 #endif
 
-        Rule *mParent = nullptr; // parent rule (NULL only for root rule)
+        Rule *mParent = nullptr; // parent rule (nullptr only for root rule)
         std::unique_ptr< QgsSymbol > mSymbol;
         double mMaximumScale = 0;
         double mMinimumScale = 0;
@@ -476,10 +509,10 @@ class CORE_EXPORT QgsRuleBasedRenderer : public QgsFeatureRenderer
     //! Returns symbol for current feature. Should not be used individually: there could be more symbols for a feature
     QgsSymbol *symbolForFeature( const QgsFeature &feature, QgsRenderContext &context ) const override;
 
-    bool renderFeature( const QgsFeature &feature, QgsRenderContext &context, int layer = -1, bool selected = false, bool drawVertexMarker = false ) override;
+    bool renderFeature( const QgsFeature &feature, QgsRenderContext &context, int layer = -1, bool selected = false, bool drawVertexMarker = false ) override SIP_THROW( QgsCsException );
 
     void startRender( QgsRenderContext &context, const QgsFields &fields ) override;
-
+    bool canSkipRender() override;
     void stopRender( QgsRenderContext &context ) override;
 
     QString filter( const QgsFields &fields = QgsFields() ) override;
@@ -490,9 +523,12 @@ class CORE_EXPORT QgsRuleBasedRenderer : public QgsFeatureRenderer
 
     QgsRuleBasedRenderer *clone() const override SIP_FACTORY;
 
-    void toSld( QDomDocument &doc, QDomElement &element, const QgsStringMap &props = QgsStringMap() ) const override;
+    void toSld( QDomDocument &doc, QDomElement &element, const QVariantMap &props = QVariantMap() ) const override;
 
-    static QgsFeatureRenderer *createFromSld( QDomElement &element, QgsWkbTypes::GeometryType geomType ) SIP_FACTORY;
+    /**
+     * Creates a new rule based renderer from an SLD XML element.
+     */
+    static QgsFeatureRenderer *createFromSld( QDomElement &element, Qgis::GeometryType geomType ) SIP_FACTORY;
 
     QgsSymbolList symbols( QgsRenderContext &context ) const override;
 
@@ -500,6 +536,7 @@ class CORE_EXPORT QgsRuleBasedRenderer : public QgsFeatureRenderer
     bool legendSymbolItemsCheckable() const override;
     bool legendSymbolItemChecked( const QString &key ) override;
     void checkLegendSymbolItem( const QString &key, bool state = true ) override;
+    QString legendKeyToExpression( const QString &key, QgsVectorLayer *layer, bool &ok ) const override;
 
     void setLegendSymbolItem( const QString &key, QgsSymbol *symbol SIP_TRANSFER ) override;
     QgsLegendSymbolList legendSymbolItems() const override;
@@ -509,6 +546,7 @@ class CORE_EXPORT QgsRuleBasedRenderer : public QgsFeatureRenderer
     QgsSymbolList originalSymbolsForFeature( const QgsFeature &feature, QgsRenderContext &context ) const override;
     QSet<QString> legendKeysForFeature( const QgsFeature &feature, QgsRenderContext &context ) const override;
     QgsFeatureRenderer::Capabilities capabilities() override { return MoreSymbolsPerFeature | Filter | ScaleDependent; }
+    bool accept( QgsStyleEntityVisitorInterface *visitor ) const override;
 
     /////
 
@@ -524,11 +562,13 @@ class CORE_EXPORT QgsRuleBasedRenderer : public QgsFeatureRenderer
     static void refineRuleScales( QgsRuleBasedRenderer::Rule *initialRule, QList<int> scales );
 
     /**
-     * creates a QgsRuleBasedRenderer from an existing renderer.
-     * \returns a new renderer if the conversion was possible, otherwise 0.
-     * \since QGIS 2.5
+     * Creates a new QgsRuleBasedRenderer from an existing \a renderer.
+     *
+     * Since QGIS 3.20, the optional \a layer parameter is required for conversions of some renderer types.
+     *
+     * \returns a new renderer if the conversion was possible, otherwise NULLPTR.
      */
-    static QgsRuleBasedRenderer *convertFromRenderer( const QgsFeatureRenderer *renderer ) SIP_FACTORY;
+    static QgsRuleBasedRenderer *convertFromRenderer( const QgsFeatureRenderer *renderer, QgsVectorLayer *layer = nullptr ) SIP_FACTORY;
 
     //! helper function to convert the size scale and rotation fields present in some other renderers to data defined symbology
     static void convertToDataDefinedSymbology( QgsSymbol *symbol, const QString &sizeScaleField, const QString &rotationField = QString() );
@@ -550,4 +590,4 @@ class CORE_EXPORT QgsRuleBasedRenderer : public QgsFeatureRenderer
 #endif
 };
 
-#endif // QGSRULEBASEDRENDERERV2_H
+#endif // QGSRULEBASEDRENDERER_H
